const gql = require('graphql-tag')
const _ = require('lodash')

const sortobject = require('deep-sort-object')
const reportTimeout = 20
const util = require('./util')

const schema = gql`
  extend type Query {
    "Introspect GraphQL server itself"
    introspect: Introspect__Query
  }

  "Introspect contains basic information about the GraphQL server itself"
  type Introspect__Query {
    "Self name"
    name: String!

    "Self version"
    version: String!

    "Self setup"
    setup: Object!

    "Self report"
    report: Object

    "Self queries"
    queries: Object

    "Self meta"
    meta: Object

    "Self markers"
    markers: Object
        
    "Self duration in ms"
    duration: Int
  }

`


const resolvers = {
  Query: {
    introspect: (_source, _args, {dataSources} ) => {
      if (dataSources.introspect) {
        return {}
      }
    },
  },
  Introspect__Query: {
    name(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getName()
    },
    version(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getVersion()
    },
    setup(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getSetup()
    },
    report(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getReport(context)
    },
    queries(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getQueries(context)
    },
    meta(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getMeta(context)
    },
    duration(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getDuration(context)
    },
    markers(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getMarkers(context)
    },
  }
}

function waitForAllOperationsDone(context) {
  return new Promise((resolve) => {
    let timeoutID
    let eventHandlerSet = false
    context.events.on('allOperationsStarted', () => {
      clearTimeout(timeoutID)
      timeoutID = null
    })
    const checkPendingOps = () => {
      if (context.pendingOps > 0) {
        if (!eventHandlerSet) {
          context.events.on('allOperationsStopped', () => {
            clearTimeout(timeoutID)
            timeoutID = null
            resolve()
          })
          eventHandlerSet = true
        }
        timeoutID = setTimeout( checkPendingOps, reportTimeout) /* ensures resolving after no reporting within time limit */
      } else {
        resolve()
      }
    }
    timeoutID = setTimeout( checkPendingOps, reportTimeout) /* ensures resolving after no reporting within time limit */
  })
}

class Introspect {
  constructor(obj) {
    this.set(obj)
  }

  set({name,version,setup}) {
    if (name) this.name = name
    if (version) this.version = version
    if (setup) this.setup = setup
  }

  getName() {
    return this.name
  }

  getVersion() {
    return this.version
  }

  getSetup() {
    return this.setup || {}
  }

  async getReport(context) {
    return waitForAllOperationsDone(context).then( () => sortobject(_.get(context,'res.introspect.report')) )
  }

  async getQueries(context) {
    return waitForAllOperationsDone(context).then( () => _.get(context,'res.introspect.queries') )
  }


  async getDuration(context) {
    return waitForAllOperationsDone(context).then( () => util.duration(context.start))
  }

  async getMarkers(context) {
    return waitForAllOperationsDone(context).then( () => sortobject(context.markers) )
  }
    
  async getMeta(context) {
    const rslt = {}
    const metaPromises = Object.keys(context.dataSources).filter( ds => !!context.dataSources[ds].query).map(ds => context.dataSources[ds].query('meta',{_id:'0'}).then(d => {
      if (d && d[0]) {
        // contentful-export-avx85g85supt-master-2020-08-27T16-24-84.json
        const pat = 'contentful-export-([^-]+)-([^-]+)-([^.]+).json'

        const match = d[0].filename.match(pat)

        const dateParts = match[3].replace('T',' ').split(/\W/)
        // const date = new Date(parseInt(dateParts[0]), parseInt(dateParts[1]) - 1, parseInt(dateParts[2].substr(0,2)),parseInt(dateParts[3]),parseInt(dateParts[4]),parseInt(dateParts[5]));
        // debug('match %y %y %y',match,dateParts,date)
        rslt[ds]={
          space: match[1], 
          environment: match[2], 
          datetime: `${dateParts[2]}-${dateParts[1]}-${dateParts[0]} ${dateParts[3]}:${dateParts[4]}:${dateParts[5]}`,
          source: d[0].filename, 
        }
      }
      return rslt[ds]
    }) )
    return await Promise.all(metaPromises).then ( da => {
      debug('da %y, %y', da, rslt)
      return rslt
    })
  }
}

module.exports = {
  schema,
  resolvers,
  datasource: Introspect
}
